from __future__ import annotations

import asyncio
import asyncio.subprocess  # disables for # https://github.com/PyCQA/pylint/issues/1469
import contextlib
import pathlib
import platform
import re
import shlex
import shutil
import tempfile
from re import Pattern
from typing import TYPE_CHECKING, ClassVar, Final

import aiohttp
import asyncstdlib
import dateutil.parser
import psutil
import rich.progress
import ujson
import yaml

from pylav._config import CONFIG_DIR
from pylav._logging import getLogger
from pylav.envvars import JAVA_EXECUTABLE
from pylav.exceptions import (
    EarlyExitError,
    IncorrectProcessFound,
    InvalidArchitectureError,
    LavalinkDownloadFailed,
    ManagedLavalinkNodeError,
    ManagedLavalinkStartFailure,
    ManagedLinkStartAbortedUseExternal,
    NodeUnhealthy,
    NoProcessFound,
    PortAlreadyInUseError,
    TooManyProcessFound,
    UnexpectedJavaResponseError,
    UnsupportedJavaError,
    WebsocketNotConnectedError,
)
from pylav.node import Node
from pylav.sql.models import NodeModel
from pylav.utils import AsyncIter, ExponentialBackoffWithReset
from pylav.vendored import aiopath

if TYPE_CHECKING:
    from pylav.client import Client

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLavPlayer", pathlib.Path(__file__))
except ImportError:
    _ = lambda x: x

LOGGER = getLogger("PyLav.ManagedNode")

LAVALINK_DOWNLOAD_DIR = CONFIG_DIR / "lavalink"
LAVALINK_DOWNLOAD_DIR = pathlib.Path(LAVALINK_DOWNLOAD_DIR)  # type: ignore
LAVALINK_DOWNLOAD_DIR.mkdir(parents=True, exist_ok=True)
LOGGER.info("Lavalink folder: %s", LAVALINK_DOWNLOAD_DIR)
_LAVALINK_JAR_FILE_FORCED_SYNC = LAVALINK_DOWNLOAD_DIR / "forced.jar"
LAVALINK_DOWNLOAD_DIR: aiopath.AsyncPath = aiopath.AsyncPath(LAVALINK_DOWNLOAD_DIR)
LAVALINK_JAR_FILE: aiopath.AsyncPath = LAVALINK_DOWNLOAD_DIR / "Lavalink.jar"
LAVALINK_JAR_FILE_FORCED: Final[aiopath.AsyncPath] = LAVALINK_DOWNLOAD_DIR / "forced.jar"
if USING_FORCED := _LAVALINK_JAR_FILE_FORCED_SYNC.exists():
    LOGGER.warning("%s found, disabling any JAR automated downloads", LAVALINK_JAR_FILE_FORCED)
    LAVALINK_JAR_FILE: aiopath.AsyncPath = LAVALINK_JAR_FILE_FORCED


LAVALINK_APP_YML: Final[aiopath.AsyncPath] = LAVALINK_DOWNLOAD_DIR / "application.yml"

_RE_READY_LINE: Final[Pattern] = re.compile(rb"Lavalink is ready to accept connections")
_FAILED_TO_START: Final[Pattern] = re.compile(rb"Web server failed to start\. (.*)")

# Version regexes
#
# We expect the output to look something like:
#     $ java -version
#     ...
#     ... version "VERSION STRING HERE" ...
#     ...
#
# There are two version formats that we might get here:
#
# - Version scheme pre JEP 223 - used by Java 8 and older
#
# examples:
# 1.8.0
# 1.8.0_275
# 1.8.0_272-b10
# 1.8.0_202-internal-201903130451-b08
# 1.8.0_272-ea-202010231715-b10
# 1.8.0_272-ea-b10
#
# Implementation based on J2SE SDK/JRE Version String Naming Convention document:
# https://www.oracle.com/java/technologies/javase/versioning-naming.html
_RE_JAVA_VERSION_LINE_PRE223 = re.compile(r'version "1\.(?P<major>[0-8])\.(?P<minor>0)(?:_\d+)?(?:-.*)?"')
# - Version scheme introduced by JEP 223 - used by Java 9 and newer
#
# examples:
# 11
# 11.0.9
# 11.0.9.1
# 11.0.9-ea
# 11.0.9-202011050024
#
# Implementation based on JEP 223 document:
# https://openjdk.java.net/jeps/223
_RE_JAVA_VERSION_LINE_223 = re.compile(r'version "(?P<major>\d+)(?:\.(?P<minor>\d+))?(?:\.\d+)*(-[a-zA-Z\d]+)?"')
_RE_SEMANTIC_VERSION_LAZY = re.compile(
    r"(?P<major>[0-9]|[1-9][0-9]*)\."
    r"(?P<minor>[0-9]|[1-9][0-9]*)\."
    r"(?P<micro>[0-9]|[1-9][0-9]*)"
    r"(?:-(?P<build>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*))?"
    r"(?:\+[0-9A-Za-z-]+)?"
)

LAVALINK_BUILD_LINE: Final[Pattern] = re.compile(rb"Build:\s+(?P<build>\d+|Unknown)")
LAVALINK_BRANCH_LINE: Final[Pattern] = re.compile(rb"Branch\s+(?P<branch>.+?)\n")
LAVALINK_JAVA_LINE: Final[Pattern] = re.compile(rb"JVM:\s+(?P<jvm>.+?)\n")
LAVALINK_LAVAPLAYER_LINE: Final[Pattern] = re.compile(rb"Lavaplayer\s+(?P<lavaplayer>.+?)\n")
LAVALINK_BUILD_TIME_LINE: Final[Pattern] = re.compile(rb"Build time:\s+(?P<build_time>.+?)\n")
LAVALINK_COMMIT_LINE: Final[Pattern] = re.compile(rb"Commit:\s+(?P<commit>.+?)\n")
LAVALINK_VERSION_LINE: Final[Pattern] = re.compile(rb"Version:\s+(?P<version>.+?)\n")
JAR_SERVER_RELEASES = "https://api.github.com/repos/freyacodes/Lavalink/releases"


def convert_function(key: str) -> str:
    return key.replace("_", "-")


def change_dict_naming_convention(data: dict) -> dict:
    new = {}
    for k, v in data.items():
        new_v = v
        if isinstance(v, dict):
            new_v = change_dict_naming_convention(v)
        elif isinstance(v, list):
            new_v = []
            for x in v:
                if isinstance(x, dict):
                    new_v.append(change_dict_naming_convention(x))
                else:
                    new_v.append(x)
        new[convert_function(k)] = new_v
    return new


def get_max_allocation_size(executable: str) -> tuple[int, bool]:
    if platform.architecture(executable)[0] == "64bit":
        max_heap_allowed = psutil.virtual_memory().total
        thinks_is_64_bit = True
    else:
        max_heap_allowed = 4 * 1024**3
        thinks_is_64_bit = False
    return max_heap_allowed, thinks_is_64_bit


class LocalNodeManager:

    _java_available: ClassVar[bool | None] = None
    _java_version: ClassVar[tuple[int, int] | None] = None
    _up_to_date: ClassVar[bool | None] = None
    _blacklisted_archs: list[str] = []

    _lavaplayer: ClassVar[str | None] = None
    _lavalink_build: ClassVar[int | None] = None
    _jvm: ClassVar[str | None] = None
    _lavalink_branch: ClassVar[str | None] = None
    _buildtime: ClassVar[str | None] = None
    _commit: ClassVar[str | None] = None
    _version: ClassVar[str | None] = None
    _java_exc: ClassVar[str] = JAVA_EXECUTABLE

    def __init__(self, client: Client, timeout: int | None = None, auto_update: bool = True) -> None:
        self._auto_update = False if USING_FORCED else auto_update
        self.ready: asyncio.Event = asyncio.Event()
        self._ci_info: dict = {"number": 0, "branchName": "", "finishDate": "", "href": "", "jar_url": ""}
        self._client = client
        self._proc: asyncio.subprocess.Process | None = None  # pylint:disable=no-member
        self._node_pid: int | None = None
        self._shutdown: bool = False
        self.start_monitor_task = None
        self.timeout = timeout
        self._args = []
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps)
        self._node_id: int = self._client.bot.user.id
        self._node: Node | None = None
        self._current_config = {}
        self._full_data: NodeModel = None  # type: ignore
        self.abort_for_unmanaged: asyncio.Event = asyncio.Event()
        self._args = []
        self._wait_for = asyncio.Event()
        self._java_path = None

    @property
    def node(self) -> Node | None:
        return self._node

    @property
    def path(self) -> str | None:
        return self._java_exc

    @property
    def jvm(self) -> str | None:
        return self._jvm

    @property
    def lavaplayer(self) -> str | None:
        return self._lavaplayer

    @property
    def ll_build(self) -> int | None:
        return self._lavalink_build

    @property
    def ll_branch(self) -> str | None:
        return self._lavalink_branch

    @property
    def build_time(self) -> str | None:
        return self._buildtime

    async def get_ci_latest_info(self) -> dict:
        async with self._client.cached_session.get(
            f"{JAR_SERVER_RELEASES}", headers={"Accept": "application/json"}
        ) as response:
            if response.status != 200:
                LOGGER.warning("Failed to get latest CI info from GitHub: %s", response.status)
                self._ci_info["number"] = -1
                return self._ci_info
            data = await response.json(loads=ujson.loads)
            release = await asyncstdlib.max(data, key=lambda x: dateutil.parser.parse(x["published_at"]))
            assets = release.get("assets", [])
            url = None
            for asset in assets:
                if asset["name"] != "Lavalink.jar":
                    continue
                url = asset.get("browser_download_url")
            date = release.get("published_at")
            branch = release.get("target_commitish")
            return {"number": int(release.get("id")), "branchName": branch, "finishDate": date, "jar_url": url}

    async def _start(self, java_path: str) -> None:
        arch_name = platform.machine()
        self._java_exc = java_path
        if arch_name in self._blacklisted_archs:
            raise InvalidArchitectureError(
                _("You are attempting to run the managed Lavalink node on an unsupported machine architecture.")
            )
        await self.process_settings()
        possible_lavalink_processes = await self.get_lavalink_process(lazy_match=True)
        if possible_lavalink_processes:
            LOGGER.info(
                "Found %s processes that match potential unmanaged Lavalink nodes.",
                len(possible_lavalink_processes),
            )
            valid_working_dirs = [
                cwd for d in possible_lavalink_processes if d.get("name") == "java" and (cwd := d.get("cwd"))
            ]
            LOGGER.debug("Found %s java processed with a cwd set.", len(valid_working_dirs))
            for cwd in valid_working_dirs:
                config = aiopath.AsyncPath(cwd) / "application.yml"
                if await config.exists() and await config.is_file():
                    LOGGER.debug(
                        "The following config file exists for an unmanaged Lavalink node %s",
                        config,
                    )
                    try:
                        async with config.open(mode="r") as config_data:
                            data = yaml.safe_load(await config_data.read())
                            data["server"]["address"]  # noqa
                            data["server"]["port"]  # noqa
                            data["lavalink"]["server"]["password"]  # noqa
                            self._node_pid = 0
                            self._current_config = data
                            raise ManagedLinkStartAbortedUseExternal
                    except ManagedLinkStartAbortedUseExternal:
                        raise
                    except Exception:
                        LOGGER.exception("Failed to read contents of %s", config)
                        continue

        await self.maybe_download_jar()
        args, msg = await self._get_jar_args()
        if msg is not None:
            LOGGER.warning(msg)
        command_string = shlex.join(args)
        LOGGER.info("Managed Lavalink node startup command: %s", command_string)
        if "-Xmx" not in command_string and msg is None:
            LOGGER.warning("Managed Lavalink node maximum allowed RAM not set or higher than available RAM")
        try:
            self._proc = await asyncio.subprocess.create_subprocess_exec(  # pylint:disable=no-member
                *args,
                cwd=str(LAVALINK_DOWNLOAD_DIR),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            self._node_pid = self._proc.pid
            LOGGER.info("Managed Lavalink node started. PID: %s", self._node_pid)
            try:
                await asyncio.wait_for(self._wait_for_launcher(), timeout=self.timeout)
            except asyncio.TimeoutError:
                LOGGER.warning("Timeout occurred whilst waiting for managed Lavalink node to be ready")
                raise
        except asyncio.TimeoutError:
            await self._partial_shutdown()
        except Exception:
            await self._partial_shutdown()
            raise

    async def process_settings(self):
        self._full_data = await self._client.node_db_manager.get_bundled_node_config()
        data = change_dict_naming_convention(self._full_data.yaml)
        # The reason this is here is to completely remove these keys from the application.yml
        # if they are set to empty values
        if not await asyncstdlib.all(
            (
                data["lavalink"]["server"]["youtubeConfig"].get("email"),
                data["lavalink"]["server"]["youtubeConfig"].get("password"),
            )
        ):
            del data["lavalink"]["server"]["youtubeConfig"]
        if not (
            data["lavalink"]["server"]["ratelimit"].get("ipBlocks")
            and data["lavalink"]["server"]["ratelimit"].get("strategy")
        ):
            del data["lavalink"]["server"]["ratelimit"]
        if data["sentry"]["dsn"]:
            data["sentry"]["tags"]["ID"] = self._client.bot.user.id
            data["sentry"]["tags"]["pylav_version"] = self._client.lib_version
        if not data["lavalink"]["server"]["httpConfig"].get("proxyHost"):
            del data["lavalink"]["server"]["httpConfig"]
        self._current_config = data
        async with LAVALINK_APP_YML.open("w") as f:
            await f.write(yaml.safe_dump(data))

    async def _get_jar_args(self) -> tuple[list[str], str | None]:
        java_available, java_version = await self._has_java()
        if not java_available:
            raise Exception(_("Pylav - Java executable not  found, tried to use: '{self._java_exc}'").format(self=self))

        java_xms, java_xmx = "64M", self._full_data.extras.get("max_ram", "2048M") if self._full_data else "2048M"

        match = re.match(r"^(\d+)([MG])$", java_xmx, flags=re.IGNORECASE)
        command_args = [self._java_exc, f"-Xms{java_xms}"]
        if (11, 0) <= java_version < (12, 0):
            command_args.append("-Djdk.tls.client.protocols=TLSv1.2")
        meta = 0, None
        invalid = None
        if (
            match
            and int(match[1]) * 1024 ** (2 if match[2].lower() == "m" else 3)
            <= (meta := get_max_allocation_size(self._java_exc))[0]
        ):
            command_args.append(f"-Xmx{java_xmx}")
        elif meta[0] is not None:
            invalid = "Managed Lavalink node RAM allocation ignored due to system limitations, please fix this."

        command_args.extend(["-jar", str(LAVALINK_JAR_FILE)])
        self._args = command_args
        return command_args, invalid

    async def _has_java(self) -> tuple[bool, tuple[int, int] | None]:
        if self._java_available:
            # Return cached value if we've checked this before
            return self._java_available, self._java_version
        java_exec = shutil.which(self._java_exc)
        java_available = java_exec is not None
        if not java_available:
            self._java_available = False
            self._java_version = None
        else:
            self._java_version = await self._get_java_version()
            self._java_available = self._java_version >= (11, 0)  # https://github.com/freyacodes/Lavalink#requirements
            self._java_exc = java_exec
        return self._java_available, self._java_version

    async def _get_java_version(self) -> tuple[int, int]:
        """This assumes we've already checked that java exists."""
        _proc: asyncio.subprocess.Process = await asyncio.create_subprocess_exec(  # pylint:disable=no-member
            self._java_exc,
            "-version",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        # java -version outputs to stderr
        _, err = await _proc.communicate()

        version_info: str = err.decode("utf-8")
        lines = version_info.splitlines()

        for line in lines:
            match = _RE_JAVA_VERSION_LINE_PRE223.search(line)
            if match is None:
                match = _RE_JAVA_VERSION_LINE_223.search(line)
            if match is None:
                match = _RE_SEMANTIC_VERSION_LAZY.search(line)
            if match is None:
                continue
            major = int(match["major"])
            minor = 0
            if minor_str := match["minor"]:
                minor = int(minor_str)

            return major, minor

        raise UnexpectedJavaResponseError(
            _("The output of `{java_exc} -version` was unexpected\n{version_info}.").format(
                java_exc=self._java_exc, version_info=version_info
            )
        )

    async def _wait_for_launcher(self) -> None:
        LOGGER.info("Waiting for Managed Lavalink node to be ready")
        async for __ in asyncstdlib.cycle("."):
            line = await self._proc.stdout.readline()
            if _RE_READY_LINE.search(line):
                self.ready.set()
                LOGGER.info("Managed Lavalink node is ready to receive requests.")
                break
            if _FAILED_TO_START.search(line):
                if f"Port {self._current_config['server']['port']} was already in use".encode() in line:
                    raise PortAlreadyInUseError(
                        _("Port {port} already in use. Managed Lavalink startup aborted.").format(
                            port=self._current_config["server"]["port"]
                        )
                    )
                raise ManagedLavalinkStartFailure(
                    _("Lavalink failed to start: {line}").format(line=line.decode("utf-8"))
                )
            if self._proc.returncode is not None:
                # Avoid Console spam only print once every 2 seconds
                raise EarlyExitError(_("Managed Lavalink node server exited early."))

    async def shutdown(self) -> None:
        if self.start_monitor_task is not None:
            self.start_monitor_task.cancel()
        if self.node:
            await self._client.node_manager.remove_node(self.node)
            self._node = None
        await self._partial_shutdown()
        await self._session.close()

    async def _partial_shutdown(self) -> None:
        self.ready.clear()
        self._wait_for.clear()
        self.abort_for_unmanaged.clear()
        # In certain situations to await self._proc.wait() is invalid so waiting on it waits forever.
        if self._shutdown is True:
            # For convenience, calling this method more than once or calling it before starting it
            # does nothing.
            return
        if self._node_pid:
            with contextlib.suppress(psutil.Error):
                p = psutil.Process(self._node_pid)
                p.terminate()
                p.kill()
        if self._proc is not None and self._proc.returncode is None:
            self._proc.terminate()
            self._proc.kill()
            await self._proc.wait()
        self._proc = None
        self._shutdown = True
        self._node_pid = None
        if self._node is not None:
            await self._client.remove_node(self._node_id)
            self._node = None

    async def _download_jar(self) -> None:
        if not self._auto_update:
            return
        LOGGER.info("Downloading Lavalink.jar...")
        jar_url = (
            self._ci_info["jar_url"] or "https://github.com/freyacodes/Lavalink/releases/download/3.5/Lavalink.jar"
        )

        async with self._session.get(jar_url, timeout=3600) as response:
            if 400 <= response.status < 600:
                raise LavalinkDownloadFailed(response=response, should_retry=True)
            fd, path = tempfile.mkstemp()
            file = open(fd, "wb")
            nbytes = 0
            with rich.progress.Progress(
                rich.progress.SpinnerColumn(),
                rich.progress.TextColumn("[progress.description]{task.description}"),
                rich.progress.BarColumn(),
                rich.progress.TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                rich.progress.TimeRemainingColumn(),
                rich.progress.TimeElapsedColumn(),
            ) as progress:
                progress_task_id = progress.add_task("[red]Downloading Lavalink.jar", total=response.content_length)
                try:
                    chunk = await response.content.read(1024)
                    while chunk:
                        chunk_size = file.write(chunk)
                        nbytes += chunk_size
                        progress.update(progress_task_id, advance=chunk_size)
                        chunk = await response.content.read(1024)
                    file.flush()
                finally:
                    file.close()

            shutil.move(path, str(LAVALINK_JAR_FILE), copy_function=shutil.copyfile)

        LOGGER.info("Successfully downloaded Lavalink.jar (%s bytes written)", format(nbytes, ","))
        self._client._config.download_id = self._ci_info["number"]

        await self._is_up_to_date()
        await self._client._config.set_download_id(self._ci_info["number"])

    async def _is_up_to_date(self):
        if self._up_to_date is True:
            # Return cached value if we've checked this before
            return True
        last_download_id = self._client._config.download_id
        args, _ = await self._get_jar_args()
        args.append("--version")
        _proc = await asyncio.subprocess.create_subprocess_exec(  # pylint:disable=no-member
            *args,
            cwd=str(LAVALINK_DOWNLOAD_DIR),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        stdout = (await _proc.communicate())[0]
        build = LAVALINK_BUILD_LINE.search(stdout) or {"build": b"Unknown"}
        branch = LAVALINK_BRANCH_LINE.search(stdout) or {"branch": b"Unknown"}
        java = LAVALINK_JAVA_LINE.search(stdout) or {"jvm": b"Unknown"}
        lavaplayer = LAVALINK_LAVAPLAYER_LINE.search(stdout) or {"lavaplayer": b"Unknown"}
        buildtime = LAVALINK_BUILD_TIME_LINE.search(stdout) or {"build_time": b"Unknown"}
        commit = LAVALINK_COMMIT_LINE.search(stdout) or {"commit": b"Unknown"}
        version = LAVALINK_VERSION_LINE.search(stdout) or {"version": b"Unknown"}
        LOGGER.info(
            "Current Lavalink meta: Lavalink build: %s, branch: %s, "
            "java: %s, lavaplayer: %s, build_time: %s commit: %s version: %s",
            build["build"],
            branch["branch"],
            java["jvm"],
            lavaplayer["lavaplayer"],
            buildtime["build_time"],
            commit["commit"],
            version["version"],
        )
        if build["build"] == b"Unknown":
            build = int(last_download_id)
        else:
            build = int(build["build"])
        date = buildtime["build_time"].decode()
        date = date.replace(".", "/")
        self._lavalink_build = build
        self._lavalink_branch = branch["branch"].decode()
        self._jvm = java["jvm"].decode()
        self._lavaplayer = lavaplayer["lavaplayer"].decode()
        self._commit = commit["commit"].decode()
        self._version = version["version"].decode()
        self._buildtime = date
        if self._auto_update:
            self._up_to_date = last_download_id == self._ci_info.get("number", -1)
        else:
            self._ci_info["number"] = build
            self._up_to_date = True
        return self._up_to_date

    async def maybe_download_jar(self):
        self._ci_info = await self.get_ci_latest_info()
        LOGGER.info("CI info: %s", self._ci_info)
        if not (await LAVALINK_JAR_FILE.exists() and await self._is_up_to_date()):
            await self._download_jar()

    async def wait_until_ready(self, timeout: float | None = None):
        tasks = [asyncio.create_task(c) for c in [self.ready.wait(), self.abort_for_unmanaged.wait()]]
        done, pending = await asyncio.wait(tasks, timeout=timeout or self.timeout, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        if done:
            done.pop().result()
        if self.abort_for_unmanaged.is_set():
            raise asyncio.TimeoutError
        if not self.ready.is_set():
            raise asyncio.TimeoutError

    async def wait_until_connected(self, timeout: float | None = None):
        tasks = [asyncio.create_task(c) for c in [self._wait_for.wait(), self.wait_until_ready()]]
        done, pending = await asyncio.wait(tasks, timeout=timeout or self.timeout, return_when=asyncio.ALL_COMPLETED)
        for task in pending:
            task.cancel()

    async def start_monitor(self, java_path: str):  # sourcery no-metrics
        retry_count = 0
        backoff = ExponentialBackoffWithReset(base=3)
        while True:
            try:
                self._shutdown = False
                if self._node_pid is None or not psutil.pid_exists(self._node_pid):
                    self.ready.clear()
                    await self._start(java_path=java_path)
                while True:
                    await self.wait_until_ready(timeout=self.timeout)
                    if not psutil.pid_exists(self._node_pid):
                        raise NoProcessFound
                    if self._node is None or not self._node.websocket.connected and not self._node.websocket.connecting:
                        await self.connect_node(reconnect=retry_count != 0, wait_for=3)
                    try:
                        node = self._client.node_manager.get_node_by_id(self._node_id)
                        if node is not None:
                            await node.wait_until_ready(timeout=30)
                        if node.websocket.connected:
                            try:
                                # Hoping this throws an exception which will then trigger a restart
                                await node.websocket.ping()
                                backoff.reset()
                                await asyncio.sleep(1)
                            except WebsocketNotConnectedError:
                                await asyncio.sleep(5)
                        elif node.websocket.connecting:
                            await node.websocket.wait_until_ready(timeout=30)
                        else:
                            raise AttributeError
                    except AttributeError as e:
                        try:
                            LOGGER.debug(
                                "Managed node monitor detected RLL is not connected to any nodes -%s", exc_info=e
                            )
                            while True:
                                node = self._client.node_manager.get_node_by_id(self._node_id)
                                if node is not None:
                                    await node.wait_until_ready(timeout=30)
                                if node and node.websocket.connected:
                                    break
                                await asyncio.sleep(1)
                        except asyncio.TimeoutError:
                            raise
                    except Exception as exc:
                        LOGGER.debug(exc, exc_info=exc)
                        raise NodeUnhealthy(str(exc)) from exc
            except (TooManyProcessFound, IncorrectProcessFound, NoProcessFound):
                await self._partial_shutdown()
            except asyncio.TimeoutError:
                delay = backoff.delay()
                await self._partial_shutdown()
                LOGGER.warning(
                    "Lavalink Managed node health check timeout, restarting in %s seconds",
                    delay,
                )
                await asyncio.sleep(delay)
            except ManagedLavalinkStartFailure:
                LOGGER.warning("Lavalink Managed node failed to start, restarting")
                await self._partial_shutdown()
                for process in await self.get_lavalink_process(
                    "-Djdk.tls.client.protocols=TLSv1.2", "-Xms64M", "-jar", cwd=str(LAVALINK_DOWNLOAD_DIR)
                ):
                    with contextlib.suppress(psutil.Error):
                        pid = process["pid"]
                        p = psutil.Process(pid)
                        p.terminate()
                        p.kill()
            except NodeUnhealthy:
                delay = backoff.delay()
                await self._partial_shutdown()
                LOGGER.warning(
                    "Lavalink Managed node health check failed, restarting in %s seconds",
                    delay,
                )
                await asyncio.sleep(delay)
            except LavalinkDownloadFailed as exc:
                delay = backoff.delay()
                if exc.should_retry:
                    LOGGER.warning(
                        "Lavalink Managed node download failed retrying in %s seconds\n%s",
                        delay,
                        exc.response,
                    )
                    retry_count += 1
                    await self._partial_shutdown()
                    await asyncio.sleep(delay)
                else:
                    LOGGER.critical(
                        "Fatal exception whilst starting managed Lavalink node, aborting...\n%s",
                        exc.response,
                    )
                    # lavalink_connection_aborted
                    return await self.shutdown()
            except InvalidArchitectureError:
                LOGGER.critical("Invalid machine architecture, cannot run a managed Lavalink node.")
                # lavalink_connection_aborted
                return await self.shutdown()
            except (UnsupportedJavaError, UnexpectedJavaResponseError) as exc:
                LOGGER.critical(exc)
                # lavalink_connection_aborted
                return await self.shutdown()
            except ManagedLinkStartAbortedUseExternal:
                LOGGER.warning("Lavalink Managed node start aborted, using the detected external Lavalink node.")
                await self.connect_node(reconnect=False, wait_for=0, external_fallback=True)
                return
            except ManagedLavalinkNodeError as exc:
                delay = backoff.delay()
                LOGGER.critical(
                    exc,
                )
                await self._partial_shutdown()
                LOGGER.warning(
                    "Lavalink Managed node startup failed retrying in %s seconds",
                    delay,
                )
                await asyncio.sleep(delay)
            except asyncio.CancelledError:
                LOGGER.warning("Lavalink Managed monitor task cancelled")
                return
            except Exception as exc:
                delay = backoff.delay()
                LOGGER.warning(
                    "Lavalink Managed node startup failed retrying in %s seconds",
                    delay,
                )
                LOGGER.info(exc, exc_info=exc)
                await self._partial_shutdown()
                await asyncio.sleep(delay)

    async def start(self, java_path: str):
        self._java_path = java_path
        if self.start_monitor_task is not None:
            await self.shutdown()
            self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps)
        self._wait_for.clear()
        self.start_monitor_task = asyncio.create_task(self.start_monitor(java_path))
        self.start_monitor_task.set_name("LavalinkManagedNode.health_monitor")

    async def connect_node(self, reconnect: bool, wait_for: float = 0.0, external_fallback: bool = False):
        # sourcery no-metrics
        await asyncio.sleep(wait_for)
        self._wait_for.clear()
        if not self.ready.is_set():
            if external_fallback:
                self.ready.set()
            else:
                raise ManagedLavalinkStartFailure()
        if reconnect:
            node = self._client.node_manager.get_node_by_id(self._node_id)
            if node is not None:
                self._node = node
                if node.websocket.connecting:
                    await node.wait_until_ready(timeout=30)
                elif node.websocket.connected:
                    LOGGER.info("Managed Lavalink node is connected")
                else:
                    LOGGER.info("Managed Lavalink node is not connected, reconnecting...")
                    await node.websocket.close()
                    await node.websocket._websocket_closed(reason="Managed Node restart")
                    await node.wait_until_ready(timeout=30)
                self._wait_for.set()
                return
        if (node := self._client.node_manager.get_node_by_id(self._node_id)) is None:
            node = self._node = await self._client.add_node(
                host=self._current_config["server"]["address"],
                port=self._current_config["server"]["port"],
                password=self._current_config["lavalink"]["server"]["password"],
                resume_key=f"ManagedNode-{self._node_pid}-{self._node_id}",
                resume_timeout=self._full_data.resume_timeout,
                name=f"PyLavPortConflictRecovery: {self._node_pid}"
                if external_fallback
                else f"{self._full_data.name}: {self._node_pid}",
                yaml=self._full_data.yaml,
                extras=self._full_data.extras,
                managed=True,
                ssl=False,
                search_only=False,
                unique_identifier=self._full_data.id,
                skip_db=True,
            )

        else:
            self._node = node
        if node.websocket.connecting:
            await node.wait_until_ready()
        elif node.websocket.connected:
            LOGGER.info("Managed Lavalink node is connected")
        else:
            LOGGER.info("Managed Lavalink node is not connected, reconnecting...")
            await node.websocket.close()
            await node.websocket._websocket_closed(reason="Managed Node restart")
            await node.wait_until_ready(timeout=30)
        self._wait_for.set()

    @staticmethod
    async def get_lavalink_process(*matches: str, cwd: str | None = None, lazy_match: bool = False):
        process_list = []
        filter_ = [cwd] if cwd else []
        async for proc in AsyncIter(psutil.process_iter()):
            with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                if cwd and await asyncio.to_thread(proc.cwd) not in filter_:
                    continue
                cmdline = await asyncio.to_thread(proc.cmdline)
                if (
                    matches
                    and await asyncstdlib.all(a in cmdline for a in matches)
                    or lazy_match
                    and await asyncstdlib.any("lavalink" in arg.lower() for arg in cmdline)
                ):
                    proc_as_dict = await asyncio.to_thread(
                        proc.as_dict, attrs=["pid", "name", "create_time", "status", "cmdline", "cwd"]
                    )

                    process_list.append(proc_as_dict)
        return process_list

    async def restart(self, java_path: str = None):
        LOGGER.info("Restarting managed Lavalink node.")
        await self.shutdown()
        await self.start(java_path=java_path or self._java_path)



from __future__ import annotations

import os

from red_commons.logging import RedTraceLogger
from red_commons.logging import getLogger as redgetLogger
from red_commons.logging import maybe_update_logger_class

maybe_update_logger_class()

LOGGER_PREFIX = os.getenv("PYLAV__LOGGER_PREFIX")

if LOGGER_PREFIX is None:
    try:
        from redbot.core.utils import AsyncIter  # noqa:

        LOGGER_PREFIX = "red."
    except ImportError:
        LOGGER_PREFIX = ""


def getLogger(name: str) -> RedTraceLogger:
    return redgetLogger(f"{LOGGER_PREFIX}{name}")



# export class hierarchy
# export iterable async file handle
from pylav.vendored.aiopath.aiopath_310.handle import IterableAIOFile
from pylav.vendored.aiopath.aiopath_310.path import (
    AsyncPath,
    AsyncPosixPath,
    AsyncPurePath,
    AsyncPurePosixPath,
    AsyncPureWindowsPath,
    AsyncWindowsPath,
)



from __future__ import annotations

import asyncio
import random
import socket

import asyncstdlib

from pylav._logging import getLogger

LOGGER = getLogger("PyLav.RadioBrowser")


class Error(Exception):
    """Base class for all exceptions raised by this module."""


class RDNSLookupError(Error):
    def __init__(self, ip):
        self.ip = ip
        self.error_msg = f"There was a problem with performing " f"reverse dns lookup for ip: {ip}"
        super().__init__(self.error_msg)


async def fetch_servers():
    """
    Get IP of all currently available `Radiob Browser` servers.
    Returns:
        list: List of IPs
    """
    ips = []
    try:
        data = await asyncio.to_thread(socket.getaddrinfo, "all.api.radio-browser.info", 80, 0, 0, socket.IPPROTO_TCP)
    except socket.gaierror:
        LOGGER.exception("Network failure")
        raise
    else:
        if data and isinstance(data[0][4], tuple):
            ips.extend(ip[4][0] for ip in data)
    return ips


async def rdns_lookup(ip: str) -> str:
    """
    Reverse DNS lookup.
    Returns:
        str: hostname
    """

    try:
        hostname, _, _ = await asyncio.to_thread(socket.gethostbyaddr, ip)
    except socket.herror as exc:
        raise RDNSLookupError(ip) from exc
    return hostname


async def fetch_hosts() -> list[str]:
    names = []
    servers = await fetch_servers()

    for ip in servers:
        try:
            host_name = await rdns_lookup(ip)
        except RDNSLookupError as exc:
            LOGGER.exception(exc.error_msg)
        else:
            names.append(host_name)
    return names


async def pick_base_url() -> str:
    hosts = await fetch_hosts()
    url = random.choice(await asyncstdlib.sorted(hosts))
    return f"https://{url}/"



from __future__ import annotations

from pylav.filters.utils import FilterMixin


class Distortion(FilterMixin):
    def __init__(
        self,
        sin_offset: float,
        sin_scale: float,
        cos_offset: float,
        cos_scale: float,
        tan_offset: float,
        tan_scale: float,
        offset: float,
        scale: float,
    ):
        self.sin_offset = sin_offset
        self.sin_scale = sin_scale
        self.cos_offset = cos_offset
        self.cos_scale = cos_scale
        self.tan_offset = tan_offset
        self.tan_scale = tan_scale
        self.offset = offset
        self.scale = scale
        self.off = False

    def to_dict(self) -> dict:
        return {
            "sin_offset": self.sin_offset,
            "sin_scale": self.sin_scale,
            "cos_offset": self.cos_offset,
            "cos_scale": self.cos_scale,
            "tan_offset": self.tan_offset,
            "tan_scale": self.tan_scale,
            "offset": self.offset,
            "scale": self.scale,
            "off": self.off,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Distortion:
        c = cls(
            sin_offset=data["sin_offset"],
            sin_scale=data["sin_scale"],
            cos_offset=data["cos_offset"],
            cos_scale=data["cos_scale"],
            tan_offset=data["tan_offset"],
            tan_scale=data["tan_scale"],
            offset=data["offset"],
            scale=data["scale"],
        )
        c.off = data["off"]
        return c

    def __repr__(self):
        return (
            "<Distortion: "
            f"sin_offset={self.sin_offset}, "
            f"sin_scale={self.sin_scale}, "
            f"cos_offset={self.cos_offset}, "
            f"cos_scale={self.cos_scale}, "
            f"tan_offset={self.tan_offset}, "
            f"tan_scale={self.tan_scale}, "
            f"offset={self.offset}, "
            f"scale={self.scale}>"
        )

    @property
    def sin_offset(self) -> float:
        return self._sin_offset

    @sin_offset.setter
    def sin_offset(self, v: float):
        self._sin_offset = v
        self.off = False

    @property
    def sin_scale(self) -> float:
        return self._sin_scale

    @sin_scale.setter
    def sin_scale(self, v: float):
        self._sin_scale = v

    @property
    def cos_offset(self) -> float:
        return self._cos_offset

    @cos_offset.setter
    def cos_offset(self, v: float):
        self._cos_offset = v
        self.off = False

    @property
    def cos_scale(self) -> float:
        return self._cos_scale

    @cos_scale.setter
    def cos_scale(self, v: float):
        self._cos_scale = v
        self.off = False

    @property
    def tan_offset(self) -> float:
        return self._tan_offset

    @tan_offset.setter
    def tan_offset(self, v: float):
        self._tan_offset = v
        self.off = False

    @property
    def tan_scale(self) -> float:
        return self._tan_scale

    @tan_scale.setter
    def tan_scale(self, v: float):
        self._tan_scale = v
        self.off = False

    @property
    def offset(self) -> float:
        return self._offset

    @offset.setter
    def offset(self, v: float):
        self._offset = v
        self.off = False

    @property
    def scale(self) -> float:
        return self._scale

    @scale.setter
    def scale(self, v: float):
        self._scale = v
        self.off = False

    @classmethod
    def default(cls) -> Distortion:
        c = cls(
            sin_offset=0.0,
            sin_scale=1.0,
            cos_offset=0.0,
            cos_scale=1.0,
            tan_offset=0.0,
            tan_scale=1.0,
            offset=0.0,
            scale=1.0,
        )
        c.off = True
        return c

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        return {
            "sinOffset": self.sin_offset,
            "sinScale": self.sin_scale,
            "cosOffset": self.cos_offset,
            "cosScale": self.cos_scale,
            "tanOffset": self.tan_offset,
            "tanScale": self.tan_scale,
            "offset": self.offset,
            "scale": self.scale,
        }

    def reset(self) -> None:
        self.sin_offset = 0.0
        self.sin_scale = 1.0
        self.cos_offset = 0.0
        self.cos_scale = 1.0
        self.tan_offset = 0.0
        self.tan_scale = 1.0
        self.offset = 0.0
        self.scale = 1.0
        self.off = True



from __future__ import annotations

from abc import ABC
from collections.abc import Awaitable, Coroutine
from typing import TYPE_CHECKING, Any, Callable, Literal, TypedDict, TypeVar, Union

import discord

if TYPE_CHECKING:
    from discord import app_commands
    from discord.ext.commands import AutoShardedBot, Bot, Cog, Context

    try:
        from redbot.core.bot import Red
        from redbot.core.bot import Red as BotClient
        from redbot.core.commands import Cog as RedCog
        from redbot.core.commands import Context as RedContext
    except ImportError:
        BotClient = Red = AutoShardedBot
        RedCog = Cog
        RedContext = Context

    from pylav.client import Client
    from pylav.query import Query
    from pylav.utils import PyLavContext

else:
    try:
        from redbot.core.bot import Red as BotClient
    except ImportError:
        from discord.ext.commands import AutoShardedBot as BotClient


_Bot = Union["Red", "Bot", "AutoShardedBot"]


class PyLavCogMixin(ABC):
    __version__: str
    bot: BotT
    lavalink: Client
    pylav: Client


T = TypeVar("T")
Coro = Coroutine[Any, Any, T]
CoroFunc = Callable[..., Coro[Any]]
MaybeCoro = Union[T, Coro[T]]
MaybeAwaitable = Union[T, Awaitable[T]]

CogT = TypeVar("CogT", bound="Optional[Union[PyLavCogMixin, RedCog, Cog]]")
Check = Callable[["ContextT"], MaybeCoro[bool]]
Hook = Union[Callable[["CogT", "ContextT"], Coro[Any]], Callable[["ContextT"], Coro[Any]]]
Error = Union[
    Callable[["CogT", "ContextT", "CommandError"], Coro[Any]],
    Callable[["ContextT", "CommandError"], Coro[Any]],
]

ContextT = TypeVar("ContextT", bound="Union[PyLavContext[Any], RedContext[Any], Context[Any]]")


class BotClientWithLavalink(BotClient):
    _pylav_client: Client
    lavalink: Client
    pylav: Client

    async def get_context(
        self, message: discord.abc.Message | InteractionT, *, cls: type[PyLavContext] = None  # noqa: F821
    ) -> PyLavContext[Any]:
        ...


class _InteractionT(discord.Interaction):
    client: BotClientWithLavalink
    response: discord.InteractionResponse
    followup: discord.Webhook
    command: app_commands.Command[Any, ..., Any] | app_commands.ContextMenu | None
    channel: discord.interactions.InteractionChannel | None


BotT = TypeVar("BotT", bound=BotClientWithLavalink, covariant=True)
InteractionT = TypeVar("InteractionT", bound="Union[_InteractionT, discord.Interaction]")
ContextObjectT = TypeVar("ContextObjectT", bound="Union[PyLavContext[Any], InteractionT[Any], Context[Any]]")

QueryT = TypeVar("QueryT", bound="Type[Query]")


class playlistInfoT(TypedDict):  # noqa
    name: str
    selectedTrack: int


class TrackInfoT(TypedDict):
    identifier: str
    isSeekable: bool
    author: str
    length: int
    title: str
    uri: str
    position: int | None
    isStream: bool
    source: str | None


class TrackT(TypedDict):
    track: str
    info: TrackInfoT


class LavalinkResponseT(TypedDict):
    loadType: Literal["TRACK_LOADED", "PLAYLIST_LOADED", "SEARCH_RESULT", "NO_MATCHES", "LOAD_FAILED"]
    playlistInfo: playlistInfoT
    tracks: list[TrackT]


class TimedFeatureT(TypedDict):
    enabled: bool
    time: int



from __future__ import annotations

from pylav.filters.utils import FilterMixin


class Timescale(FilterMixin):
    def __init__(self, speed: float, pitch: float, rate: float):
        self.speed = speed
        self.pitch = pitch
        self.rate = rate
        self.off = False

    def to_dict(self) -> dict:
        return {
            "speed": self.speed,
            "pitch": self.pitch,
            "rate": self.rate,
            "off": self.off,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Timescale:
        c = cls(speed=data["speed"], pitch=data["pitch"], rate=data["rate"])
        c.off = data["off"]
        return c

    def __repr__(self):
        return f"<Timescale: speed={self.speed}, pitch={self.pitch}, rate={self.rate}>"

    @property
    def speed(self) -> float:
        return self._speed

    @speed.setter
    def speed(self, v: float):
        self._speed = v
        self.off = False

    @property
    def pitch(self) -> float:
        return self._pitch

    @pitch.setter
    def pitch(self, v: float):
        self._pitch = v
        self.off = False

    @property
    def rate(self) -> float:
        return self._rate

    @rate.setter
    def rate(self, v: float):
        self._rate = v
        self.off = False

    @classmethod
    def default(cls) -> Timescale:
        c = cls(speed=1.0, pitch=1.0, rate=1.0)
        c.off = True
        return c

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        return {
            "speed": self.speed,
            "pitch": self.pitch,
            "rate": self.rate,
        }

    def reset(self) -> None:
        self.speed = 1.0
        self.pitch = 1.0
        self.rate = 1.0
        self.off = True



from __future__ import annotations

import collections
from typing import Final

from deepdiff import DeepDiff

from pylav.filters.utils import FilterMixin


class Equalizer(FilterMixin):
    """Class representing a usable equalizer.

    Parameters
    ------------
    levels: List[Tuple[int, float]]
        A list of tuple pairs containing a band int and gain float.
    name: str
        An Optional string to name this Equalizer. Defaults to 'CustomEqualizer'
    """

    def __init__(self, *, levels: list, name: str = "CustomEqualizer"):
        self.band_count: Final[int] = 15
        self._eq = self._factory(levels)
        self._raw = levels
        self._name = name

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the Equalizer"""
        return {"equalizer": self._eq, "name": self._name, "off": self.off}

    @classmethod
    def from_dict(cls, data: dict) -> Equalizer:
        """Creates an Equalizer from a dictionary"""
        c = cls(levels=data["equalizer"], name=data["name"])
        c.off = data["off"]
        return c

    def __str__(self):
        return self._name

    def __repr__(self):
        return f"<Equalizer: name={self._name}, eq={self._eq}>"

    @property
    def index(self):
        return {d["band"]: dict(d, index=index) for (index, d) in enumerate(self._eq)}

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Equalizer):
            return bool(DeepDiff(self._eq, other._eq, ignore_order=True))
        return NotImplemented

    @property
    def name(self) -> str:
        """The Equalizers friendly name."""
        return self._name

    def _factory(self, levels: list) -> list[dict[str, int | float]]:
        if not levels:
            return self.default()._eq
        if not isinstance(levels[0], dict):
            raise TypeError("Equalizer levels should be a list of dictionaries.")

        _dict = collections.defaultdict(float)
        _dict.update(d.values() for d in levels)
        _dict = [{"band": i, "gain": _dict[i]} for i in range(self.band_count)]

        return _dict

    @classmethod
    def build(cls, *, levels: list, name: str = "CustomEqualizer") -> Equalizer:
        """Build a custom Equalizer class with the provided levels.

        Parameters
        ------------
        levels: List[Dict[str, Union[int, float]]]
            A list of dictionaries containing the band and gain for each band.
        name: str
            An Optional string to name this Equalizer. Defaults to 'CustomEqualizer'
        """
        return cls(levels=levels, name=name)

    @classmethod
    def flat(cls) -> Equalizer:
        """Flat Equalizer.
        Resets your EQ to Flat.
        """
        return cls(
            levels=[
                {"band": 0, "gain": 0.0},
                {"band": 1, "gain": 0.0},
                {"band": 2, "gain": 0.0},
                {"band": 3, "gain": 0.0},
                {"band": 4, "gain": 0.0},
                {"band": 5, "gain": 0.0},
                {"band": 6, "gain": 0.0},
                {"band": 7, "gain": 0.0},
                {"band": 8, "gain": 0.0},
                {"band": 9, "gain": 0.0},
                {"band": 10, "gain": 0.0},
                {"band": 11, "gain": 0.0},
                {"band": 12, "gain": 0.0},
                {"band": 13, "gain": 0.0},
                {"band": 14, "gain": 0.0},
            ],
            name="Default",
        )

    @classmethod
    def default(cls) -> Equalizer:
        return cls.flat()

    def get(self) -> list[dict[str, int | float]]:
        return [] if self.off else self._eq

    def reset(self) -> None:
        self.off = True
        eq = Equalizer.flat()
        self._eq = eq._eq
        self._name = eq._name
        self._raw = eq._raw

    def set_gain(self, band: int, gain: float) -> None:
        if band < 0 or band >= self.band_count:
            raise IndexError(f"Band {band} does not exist!")

        gain = float(min(max(gain, -0.25), 1.0))
        band = next((index for (index, d) in enumerate(self._eq) if d["band"] == band), None)
        self._eq[band]["gain"] = gain

    def get_gain(self, band: int) -> float:
        if band < 0 or band >= self.band_count:
            raise IndexError(f"Band {band} does not exist!")
        return self.index[band].get("gain", 0.0)

    def visualise(self):
        block = ""
        bands = [str(band + 1).zfill(2) for band in range(self.band_count)]
        bottom = (" " * 8) + " ".join(bands)
        gains = [x * 0.01 for x in range(-25, 105, 5)]
        gains.reverse()

        for gain in gains:
            prefix = ""
            if gain > 0:
                prefix = "+"
            elif gain == 0:
                prefix = " "

            block += f"{prefix}{gain:.2f} | "

            for value in self._eq:
                cur_gain = value.get("gain", 0.0)
                block += "[] " if cur_gain >= gain else "   "
            block += "\n"

        block += bottom
        return block



from __future__ import annotations

import asyncio
import contextlib
import datetime
from typing import TYPE_CHECKING

from pylav._logging import getLogger
from pylav.exceptions import EntryNotFoundError
from pylav.sql import tables
from pylav.sql.models import QueryModel
from pylav.utils import AsyncIter

if TYPE_CHECKING:
    from pylav.client import Client
    from pylav.query import Query

LOGGER = getLogger("PyLav.QueryCacheManager")


class QueryCacheManager:
    def __init__(self, client: Client):
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    async def get_query(query: Query) -> QueryModel | None:
        if query.is_local or query.is_custom_playlist or query.is_http:
            # Do not cache local queries and single track urls or http source entries
            return None
        query = (
            await tables.QueryRow.select()
            .output(as_json=True)
            .where(
                (tables.QueryRow.identifier == query.query_identifier)
                & (
                    tables.QueryRow.last_updated
                    > datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(days=30)
                )
            )
            .first()
        )
        if query:
            return QueryModel(**query)

    @staticmethod
    async def add_query(query: Query, result: dict):
        if query.is_local or query.is_custom_playlist or query.is_http:
            # Do not cache local queries and single track urls or http source entries
            return
        if result.get("loadType") in ["NO_MATCHES", "LOAD_FAILED", None]:
            return

        tracks = result.get("tracks", [])
        if not tracks:
            raise EntryNotFoundError(f"No tracks found for query {query.query_identifier}")
        name = result.get("playlistInfo", {}).get("name", None)
        await QueryModel(
            identifier=query.query_identifier,
            name=name,
            tracks=[t["track"] async for t in AsyncIter(tracks)],
            last_updated=datetime.datetime.now(tz=datetime.timezone.utc),
        ).save()

    @staticmethod
    async def delete_old() -> None:
        with contextlib.suppress(
            asyncio.exceptions.CancelledError,
        ):
            LOGGER.trace("Deleting old queries")
            await tables.QueryRow.delete().where(
                tables.QueryRow.last_updated
                <= (datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(days=30))
            )
            LOGGER.trace("Deleted old queries")

    @staticmethod
    async def wipe() -> None:
        LOGGER.trace("Wiping query cache")
        await tables.QueryRow.delete(force=True)
        LOGGER.trace("Wiped query cache")

    @staticmethod
    async def delete_older_than(days: int) -> None:
        await tables.QueryRow.delete().where(
            tables.QueryRow.last_updated
            <= (datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(days=days))
        )

    @staticmethod
    async def delete_query(query: Query) -> None:
        await tables.QueryRow.delete().where(tables.QueryRow.identifier == query.query_identifier)

    @staticmethod
    async def size() -> int:
        return await tables.QueryRow.count()



from __future__ import annotations

from pylav.filters.utils import FilterMixin


class LowPass(FilterMixin):
    def __init__(self, smoothing: float):
        self._smoothing = smoothing
        self.off = False

    def to_dict(self) -> dict:
        return {
            "smoothing": self.smoothing,
            "off": self.off,
        }

    @classmethod
    def from_dict(cls, data: dict) -> LowPass:
        c = cls(
            smoothing=data["smoothing"],
        )
        c.off = data["off"]
        return c

    def __repr__(self):
        return f"<LowPass: smoothing={self.smoothing}>"

    @property
    def smoothing(self) -> float:
        return self._smoothing

    @smoothing.setter
    def smoothing(self, v: float):
        self._smoothing = v
        self.off = False

    @classmethod
    def default(cls) -> LowPass:
        c = cls(smoothing=20.0)
        c.off = True
        return c

    def get(self) -> dict[str, float]:
        return (
            {}
            if self.off
            else {
                "smoothing": self.smoothing,
            }
        )

    def reset(self) -> None:
        self.smoothing = 20.0
        self.off = True



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav._logging import getLogger
from pylav.exceptions import EntryNotFoundError
from pylav.sql import tables
from pylav.sql.models import NodeModel
from pylav.utils.built_in_node import NODE_DEFAULT_SETTINGS

if TYPE_CHECKING:
    from pylav.client import Client

LOGGER = getLogger("PyLav.NodeConfigManager")


class NodeConfigManager:
    def __init__(self, client: Client):
        self._client = client
        self.currently_in_db = {
            self._client.bot.user.id,
        }

    @property
    def client(self) -> Client:
        return self._client

    @property
    async def bundled_node_config(self) -> NodeModel:
        return await self.get_bundled_node_config()

    async def get_node_config(self, node_id: int) -> NodeModel:
        node = await tables.NodeRow.select().output(load_json=True).where(tables.NodeRow.id == node_id).first()
        if not node:
            raise EntryNotFoundError(f"Node with id {node_id} not found")
        model = NodeModel(**node)
        self.currently_in_db.add(model.id)
        return model

    async def get_all_unamanaged_nodes(self, dedupe: bool = True) -> list[NodeModel]:
        model_list = [
            NodeModel(**node)
            for node in await tables.NodeRow.select()
            .output(load_json=True)
            .where(tables.NodeRow.managed == False)  # noqa: E712
        ]
        new_model_list = list(set(model_list)) if dedupe else model_list
        for n in new_model_list:
            self.currently_in_db.add(n.id)
        return new_model_list

    async def get_all_nodes(self) -> list[NodeModel]:
        model_list = [
            NodeModel(**node)
            for node in await tables.NodeRow.select().output(load_json=True).where(tables.NodeRow.managed == False)
        ]
        for n in model_list:
            self.currently_in_db.add(n.id)
        model_list.append(await self.get_bundled_node_config())
        return model_list

    async def get_bundled_node_config(self) -> NodeModel:
        node = (
            await tables.NodeRow.objects()
            .output(load_json=True)
            .get_or_create(
                (tables.NodeRow.id == self._client.bot.user.id) & (tables.NodeRow.managed == True),  # noqa: E712
                defaults={
                    tables.NodeRow.ssl: False,
                    tables.NodeRow.reconnect_attempts: -1,
                    tables.NodeRow.search_only: False,
                    tables.NodeRow.yaml: NODE_DEFAULT_SETTINGS,
                    tables.NodeRow.name: "PyLavManagedNode",
                    tables.NodeRow.managed: True,
                    tables.NodeRow.resume_key: None,
                    tables.NodeRow.resume_timeout: 600,
                    tables.NodeRow.extras: {"max_ram": "2048M"},
                },
            )
        )
        data = node.to_dict()
        if "max_ram" not in data["extras"]:
            data["extras"]["max_ram"] = "2048M"
        return NodeModel(**data)

    async def add_node(
        self,
        host: str,
        port: int,
        password: str,
        unique_identifier: int,
        name: str,
        resume_key: str = None,
        resume_timeout: int = 60,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        extras: dict = None,
        disabled_sources: list[str] = None,
        yaml: dict = None,
    ) -> NodeModel:
        """
        Add a new node to the database.

        Parameters
        ----------
        host: str
            The host of the node.
        port: int
            The port of the node.
        password: str
            The password of the node.
        unique_identifier: int
            The unique identifier of the node.
        name: str
            The name of the node.
        resume_key: str
            The resume key of the node.
        resume_timeout: int
            The resume timeout of the node.
        reconnect_attempts: int
            The reconnect attempts of the node.
        ssl: bool
            Whether the node is using ssl.
        search_only: bool
            Whether the node is search only.
        managed: bool
            Whether the node is managed.
        extras: dict
            The extras of the node.
        disabled_sources: list[str]
            The disabled sources of the node.
        yaml: dict
            The yaml of the node.
        """
        if unique_identifier in self.currently_in_db:
            return await self.get_node_config(unique_identifier)

        data = dict(
            yaml=yaml or {"server": {}, "lavalink": {"server": {}}},
            id=unique_identifier,
            ssl=ssl,
            reconnect_attempts=reconnect_attempts,
            search_only=search_only,
            resume_key=resume_key,
            resume_timeout=resume_timeout,
            managed=managed,
            name=name,
            disabled_sources=disabled_sources or [],
            extras=extras or {},
        )
        data["yaml"]["server"]["address"] = host  # type: ignore
        data["yaml"]["server"]["port"] = port  # type: ignore
        data["yaml"]["lavalink"]["server"]["password"] = password

        node = NodeModel(**data)
        await node.save()
        self.currently_in_db.add(node.id)
        return node

    async def update_node(
        self,
        host: str,
        port: int,
        password: str,
        unique_identifier: int,
        resume_key: str = None,
        resume_timeout: int = 60,
        name: str = None,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        managed: bool = False,
        extras: dict = None,
        yaml: dict = None,
        disabled_sources: list[str] = None,
    ) -> NodeModel:
        data = dict(
            yaml=yaml or {"server": {}, "lavalink": {"server": {}}},
            id=unique_identifier,
            ssl=ssl,
            reconnect_attempts=reconnect_attempts,
            search_only=search_only,
            resume_key=resume_key,
            resume_timeout=resume_timeout,
            managed=managed,
            name=name,
            disabled_sources=disabled_sources or [],
            extras=extras or {},
        )
        data["yaml"]["server"]["address"] = host  # type: ignore
        data["yaml"]["server"]["port"] = port  # type: ignore
        data["yaml"]["lavalink"]["server"]["password"] = password
        node = NodeModel(**data)
        await node.save()
        self.currently_in_db.add(node.id)
        return node

    async def delete(self, node_id: int) -> None:
        if node_id == self._client.bot.user.id:
            raise ValueError("Cannot delete bundled node")
        await tables.NodeRow.delete().where(
            (tables.NodeRow.id == node_id) & (tables.NodeRow.managed != True)
        )  # noqa: E712
        self.currently_in_db.discard(node_id)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import asyncstdlib
from discord.app_commands import Choice, Transformer
from discord.ext import commands

from pylav.node import Node
from pylav.types import ContextT, InteractionT

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLavPlayer", Path(__file__))
except ImportError:
    _ = lambda x: x

if TYPE_CHECKING:

    NodeConverter = TypeVar("NodeConverter", bound=list[Node])
else:

    class NodeConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: ContextT, arg: str) -> list[Node]:
            """Converts a node name or ID to a list of matching objects."""
            from pylav import EntryNotFoundError

            try:
                nodes = ctx.lavalink.node_manager.nodes
            except EntryNotFoundError as e:
                raise commands.BadArgument(_("Node with name or id `{arg}` not found.").format(arg=arg)) from e
            if r := await asyncstdlib.list(
                asyncstdlib.filter(lambda n: arg.lower() in n.name.lower() or arg == f"{n.identifier}", nodes)
            ):
                return r
            raise commands.BadArgument(_("Node with name or id `{arg}` not found.").format(arg=arg))

        @classmethod
        async def transform(cls, interaction: InteractionT, argument: str) -> list[Node]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: InteractionT, current: str) -> list[Choice]:
            nodes = interaction.client.lavalink.node_manager.nodes

            return [
                Choice(name=n.name[:99], value=f"{n.identifier}") for n in nodes if current.lower() in n.name.lower()
            ][:25]



from __future__ import annotations

from pylav._city_dump import US_CITY_DUMP

REGION_TO_COUNTRY_COORDINATE_MAPPING = {}
REGION_TO_COUNTRY_COORDINATE_MAPPING |= US_CITY_DUMP
REGION_TO_COUNTRY_COORDINATE_MAPPING |= {
    "hongkong": (22.2793278, 114.1628131),
    "singapore": (1.357107, 103.8194992),
    "sydney": (-33.8698439, 151.2082848),
    "seattle": (47.608013, -122.335167),
    "japan": (36.5748441, 139.2394179),
    "southafrica": (-28.8166236, 24.991639),
    "india": (22.3511148, 78.6677428),
    "eu": (46.603354, 1.8883335),
    "amsterdam": (52.3727598, 4.8936041),
    "frankfurt": (50.1106444, 8.6820917),
    "russia": (64.6863136, 97.7453061),
    "london": (51.5073219, -0.1276474),
    "us_central": (41.7872548, -87.8410043),
    "us_west": (37.7577627, -122.4727051),
    "us_east": (40.707938, -74.0423759),
    "us_south": (32.7870795, -96.7988588),
    "brazil": (-10.3333333, -53.2),
    "rotterdam": (51.9240069, 4.4777325),
    "santa_clara": (37.3541079, -121.9552368),
    "unknown_pylav": (0, 0),
}

DEFAULT_REGIONS = list(REGION_TO_COUNTRY_COORDINATE_MAPPING.keys())

SUPPORTED_SOURCES = {
    "youtube",
    "soundcloud",
    "twitch",
    "bandcamp",
    "vimeo",
    "http",
    "getyarn",
    "clypit",
    "speak",
    "pornhub",
    "reddit",
    "ocremix",
    "tiktok",
    "mixcloud",
    "soundgasm",
    "gcloud-tts",
    "sponsorblock",
    "local",
    "spotify",
    "applemusic",
}

BUNDLED_PLAYLIST_IDS = {1, 2, 1000001, 1000002, 1000003, 1000004}
BUNDLED_NODES_IDS = {1, 2, 1001}

# Mapping of the PyLav public nodes
#   The floats entries are possible locations but not yet being used.
PYLAV_NODES = {
    1: ("london", (1.3213765, 103.6956208)),
    2: ("new_york_city", (40.606872, -74.1769477)),
}

PYLAV_NODE_SETTINGS = {
    "port": 443,
    "ssl": True,
    "password": "default",
    "resume_timeout": 600,
    "reconnect_attempts": -1,
    "search_only": False,
    "managed": False,
    "disabled_sources": ["local", "http"],
    "skip_db": True,
}

PYLAV_BUNDLED_NODES_SETTINGS = {
    "ll-gb.draper.wtf": dict(
        **PYLAV_NODE_SETTINGS, host="ll-gb.draper.wtf", unique_identifier=1, name="PyLav London (Bundled)"
    ),
    "ll-us-ny.draper.wtf": dict(
        **PYLAV_NODE_SETTINGS, host="ll-us-ny.draper.wtf", unique_identifier=2, name="PyLav US-NY (Bundled)"
    ),
    "lava.link": {
        "host": "lava.link",
        "unique_identifier": 1001,
        "name": "Lava.Link (Bundled)",
        "port": 80,
        "resume_timeout": 600,
        "reconnect_attempts": -1,
        "ssl": False,
        "search_only": False,
        "managed": False,
        "disabled_sources": [
            "clypit",
            "reddit",
            "local",
            "tiktok",
            "speak",
            "pornhub",
            "soundgasm",
            "applemusic",
            "mixcloud",
            "sponsorblock",
            "getyarn",
            "spotify",
            "gcloud-tts",
            "ocremix",
        ],
        "skip_db": True,
    },
}



from __future__ import annotations

import asyncio
import os

from pylav.m3u8_parser.httpclient import DefaultHTTPClient, _parsed_url
from pylav.m3u8_parser.model import (
    M3U8,
    ContentSteering,
    DateRange,
    DateRangeList,
    IFramePlaylist,
    Key,
    Media,
    MediaList,
    PartialSegment,
    PartialSegmentList,
    PartInformation,
    Playlist,
    PlaylistList,
    PreloadHint,
    RenditionReport,
    RenditionReportList,
    Segment,
    SegmentList,
    ServerControl,
    Skip,
    Start,
)
from pylav.m3u8_parser.parser import ParseError, is_url, parse
from pylav.vendored import aiopath

# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.


__all__ = (
    "M3U8",
    "Segment",
    "SegmentList",
    "PartialSegment",
    "PartialSegmentList",
    "Key",
    "Playlist",
    "IFramePlaylist",
    "Media",
    "MediaList",
    "PlaylistList",
    "Start",
    "RenditionReport",
    "RenditionReportList",
    "ServerControl",
    "Skip",
    "PartInformation",
    "PreloadHint",
    "DateRange",
    "DateRangeList",
    "ContentSteering",
    "loads",
    "load",
    "parse",
    "ParseError",
)


async def loads(self, content: str, uri: str = None, custom_tags_parser=None):
    """
    Given a string with a m3u8 content, returns a M3U8 object.
    Optionally parses a uri to set a correct base_uri on the M3U8 object.
    Raises ValueError if invalid content
    """

    if uri is None:
        return await asyncio.to_thread(M3U8, content, custom_tags_parser=custom_tags_parser)
    else:
        _parsed_url(uri)


async def load(
    self,
    uri: str,
    timeout: float = None,
    headers: dict = None,
    custom_tags_parser=None,
    http_client: type[DefaultHTTPClient] = DefaultHTTPClient(),  # type: ignore
    verify_ssl: bool = True,
):
    """
    Retrieves the content from a given URI and returns a M3U8 object.
    Raises ValueError if invalid content or IOError if request fails.
    """
    if headers is None:
        headers = {}
    if is_url(uri):
        content, base_uri = await http_client.download(uri=uri, timeout=timeout, headers=headers, verify_ssl=verify_ssl)
        return await asyncio.to_thread(M3U8, content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)
    else:
        return await _load_from_file(uri, custom_tags_parser)


async def _load_from_file(uri: str, custom_tags_parser=None):
    """Loads a m3u8 file from the filesystem

    Parameters:
    -----------
    uri: str
        uri of the m3u8 file
    custom_tags_parser: callable
        custom tags parser function
    """

    path = aiopath.AsyncPath(uri)
    async with path.open("r", encoding="utf8") as file:
        raw_content = (await file.read()).strip()
    base_uri = os.path.dirname(uri)
    return M3U8(raw_content, base_uri=base_uri, custom_tags_parser=custom_tags_parser)



from pylav._config import update_event_loop_policy  # isort:skip
from pylav._logging import getLogger  # isort:skip
from pylav import converters
from pylav.client import Client
from pylav.exceptions import *
from pylav.player import Player
from pylav.query import Query
from pylav.tracks import Track

__all__ = (
    "PyLavError",
    "NodeError",
    "WebsocketNotConnectedError",
    "TrackError",
    "ManagedLavalinkNodeError",
    "HTTPError",
    "Unauthorized",
    "InvalidTrack",
    "NodeUnhealthy",
    "InvalidArchitectureError",
    "ManagedLavalinkAlreadyRunningError",
    "PortAlreadyInUseError",
    "ManagedLavalinkStartFailure",
    "ManagedLavalinkPreviouslyShutdownError",
    "EarlyExitError",
    "UnsupportedJavaError",
    "UnexpectedJavaResponseError",
    "NoProcessFound",
    "IncorrectProcessFound",
    "TooManyProcessFound",
    "LavalinkDownloadFailed",
    "TrackNotFound",
    "CogAlreadyRegistered",
    "CogHasBeenRegistered",
    "AnotherClientAlreadyRegistered",
    "ManagedLinkStartAbortedUseExternal",
    "NoNodeAvailable",
    "SQLError",
    "EntryNotFoundError",
    "PyLavNotInitialized",
    "PlaylistError",
    "InvalidPlaylist",
    "Track",
    "Player",
    "Client",
    "Query",
    "converters",
    "update_event_loop_policy",
)



import contextlib
from collections.abc import AsyncIterable

import asyncpg
from aiohttp_client_cache import BaseCache, CacheBackend, ResponseOrKey
from aiohttp_client_cache.docs import extend_init_signature

from pylav.sql import tables


def postgres_template():
    pass


@extend_init_signature(CacheBackend, postgres_template)
class PostgresCacheBackend(CacheBackend):
    """Wrapper for higher-level cache operations. In most cases, the only thing you need to specify here is which storage class(es) to use."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.redirects = PostgresStorage(**kwargs)
        self.responses = PostgresStorage(**kwargs)


class PostgresStorage(BaseCache):
    """interface for lower-level backend storage operations"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def contains(self, key: str) -> bool:
        """Check if a key is stored in the cache"""
        response = await tables.AioHttpCacheRow.select().where(tables.AioHttpCacheRow.key == key).limit(1)
        return bool(response)

    async def clear(self) -> None:
        """Delete all items from the cache"""
        await tables.AioHttpCacheRow.delete(force=True)

    async def delete(self, key: str) -> None:
        """Delete an item from the cache"""
        await tables.AioHttpCacheRow.delete().where(tables.AioHttpCacheRow.key == key)

    async def keys(self) -> AsyncIterable[str]:
        """Get all keys stored in the cache"""
        async with await tables.AioHttpCacheRow.select(tables.AioHttpCacheRow.key).batch(batch_size=10) as batch:
            async for _batch in batch:
                yield _batch["key"]

    async def read(self, key: str) -> ResponseOrKey:
        """Read an item from the cache"""
        response = (
            await tables.AioHttpCacheRow.select(tables.AioHttpCacheRow.value)
            .where(tables.AioHttpCacheRow.key == key)
            .limit(1)
        )
        if not response:
            return None
        return self.deserialize(response["value"])

    async def size(self) -> int:
        """Get the number of items in the cache"""
        return await tables.AioHttpCacheRow.count()

    def values(self) -> AsyncIterable[ResponseOrKey]:
        """Get all values stored in the cache"""
        return self._values()

    async def _values(self) -> AsyncIterable[ResponseOrKey]:
        async with await tables.AioHttpCacheRow.select(tables.AioHttpCacheRow.value).batch(batch_size=10) as batch:
            async for _batch in batch:
                yield self.deserialize(_batch["value"])

    async def write(self, key: str, item: ResponseOrKey):
        """Write an item to the cache"""

        values = {tables.AioHttpCacheRow.value: self.serialize(item)}
        while True:
            with contextlib.suppress(asyncpg.exceptions.UniqueViolationError):
                entry = await tables.AioHttpCacheRow.objects().get_or_create(
                    tables.AioHttpCacheRow.key == key, defaults=values
                )
                break
        if not entry._was_created:
            await tables.AioHttpCacheRow.update(values).where(tables.AioHttpCacheRow.key == key)

    async def bulk_delete(self, keys: set[str]) -> None:
        """Delete multiple items from the cache"""
        await tables.AioHttpCacheRow.delete().where(tables.AioHttpCacheRow.key.is_in(list(keys)))



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.events import (
    Event,
    FiltersAppliedEvent,
    NodeChangedEvent,
    NodeConnectedEvent,
    NodeDisconnectedEvent,
    PlayerConnectedEvent,
    PlayerDisconnectedEvent,
    PlayerMovedEvent,
    PlayerPausedEvent,
    PlayerRepeatEvent,
    PlayerRestoredEvent,
    PlayerResumedEvent,
    PlayerStoppedEvent,
    PlayerUpdateEvent,
    PlayerVolumeChangedEvent,
    QueueEndEvent,
    QueueShuffledEvent,
    QueueTrackPositionChangedEvent,
    QueueTracksRemovedEvent,
    SegmentSkippedEvent,
    SegmentsLoadedEvent,
    TrackAutoPlayEvent,
    TrackEndEvent,
    TrackExceptionEvent,
    TrackPreviousRequestedEvent,
    TrackResumedEvent,
    TrackSeekEvent,
    TrackSkippedEvent,
    TracksRequestedEvent,
    TrackStartAppleMusicEvent,
    TrackStartBandcampEvent,
    TrackStartClypitEvent,
    TrackStartEvent,
    TrackStartGCTTSEvent,
    TrackStartGetYarnEvent,
    TrackStartHTTPEvent,
    TrackStartLocalFileEvent,
    TrackStartMixCloudEvent,
    TrackStartNicoNicoEvent,
    TrackStartOCRMixEvent,
    TrackStartPornHubEvent,
    TrackStartRedditEvent,
    TrackStartSoundCloudEvent,
    TrackStartSoundgasmEvent,
    TrackStartSpeakEvent,
    TrackStartSpotifyEvent,
    TrackStartTikTokEvent,
    TrackStartTwitchEvent,
    TrackStartVimeoEvent,
    TrackStartYouTubeEvent,
    TrackStartYouTubeMusicEvent,
    TrackStuckEvent,
    WebSocketClosedEvent,
)

if TYPE_CHECKING:
    from pylav.client import Client


class DispatchManager:
    """
    The Dispatcher is responsible for dispatching events to the appropriate
    handlers.

    The method names are the event names.

    You can listen to events by adding the following to your client:


    Examples
    --------
    >>> from discord.app_commands import commands

    >>> @commands.Cog.listener()
    >>> async def on_pylav_queue_ended(self, event: QueueEndEvent):
    >>>    print(f"Queue ended: {event.player}")

    >>> @commands.Cog.listener()
    >>> async def on_pylav_track_stuck(self, event: TrackStuckEvent):
    >>>    print(f"Track got stuck: {event.track}")

    """

    def __init__(self, client: Client):
        self._client = client
        self.dispatcher = client.bot.dispatch
        self.mapping = {
            TrackStuckEvent: "pylav_track_stuck",
            TrackExceptionEvent: "pylav_track_exception",
            TrackEndEvent: "pylav_track_end",
            TrackStartEvent: "pylav_track_start",
            TrackStartYouTubeMusicEvent: "pylav_track_start_youtube_music",
            TrackStartSpotifyEvent: "pylav_track_start_spotify",
            TrackStartAppleMusicEvent: "pylav_track_start_apple_music",
            TrackStartLocalFileEvent: "pylav_track_start_localfile",
            TrackStartHTTPEvent: "pylav_track_start_http",
            TrackStartSpeakEvent: "pylav_track_start_speak",
            TrackStartYouTubeEvent: "pylav_track_start_youtube",
            TrackStartClypitEvent: "pylav_track_start_clypit",
            TrackStartGetYarnEvent: "pylav_track_start_getyarn",
            TrackStartMixCloudEvent: "pylav_track_start_mixcloud",
            TrackStartOCRMixEvent: "pylav_track_start_ocrmix",
            TrackStartPornHubEvent: "pylav_track_start_pornhub",
            TrackStartRedditEvent: "pylav_track_start_reddit",
            TrackStartSoundgasmEvent: "pylav_track_start_soundgasm",
            TrackStartTikTokEvent: "pylav_track_start_tiktok",
            TrackStartBandcampEvent: "pylav_track_start_bandcamp",
            TrackStartSoundCloudEvent: "pylav_track_start_soundcloud",
            TrackStartTwitchEvent: "pylav_track_start_twitch",
            TrackStartVimeoEvent: "pylav_track_start_vimeo",
            TrackStartGCTTSEvent: "pylav_track_start_gctts",
            TrackStartNicoNicoEvent: "pylav_track_start_niconico",
            TrackSkippedEvent: "pylav_track_skipped",
            TrackSeekEvent: "pylav_track_seek",
            TrackPreviousRequestedEvent: "pylav_track_previous_requested",
            TracksRequestedEvent: "pylav_tracks_requested",
            TrackAutoPlayEvent: "pylav_track_autoplay",
            TrackResumedEvent: "pylav_track_resumed",
            QueueShuffledEvent: "pylav_queue_shuffled",
            QueueEndEvent: "pylav_queue_end",
            QueueTrackPositionChangedEvent: "pylav_queue_track_position_changed",
            QueueTracksRemovedEvent: "pylav_queue_tracks_removed",
            PlayerUpdateEvent: "pylav_player_update",
            PlayerPausedEvent: "pylav_player_paused",
            PlayerStoppedEvent: "pylav_player_stopped",
            PlayerResumedEvent: "pylav_player_resumed",
            PlayerMovedEvent: "pylav_player_moved",
            PlayerDisconnectedEvent: "pylav_player_disconnected",
            PlayerConnectedEvent: "pylav_player_connected",
            PlayerVolumeChangedEvent: "pylav_player_volume_changed",
            PlayerRepeatEvent: "pylav_player_repeat",
            PlayerRestoredEvent: "pylav_player_restored",
            SegmentSkippedEvent: "pylav_segment_skipped",
            SegmentsLoadedEvent: "pylav_segments_loaded",
            FiltersAppliedEvent: "pylav_filters_applied",
            NodeConnectedEvent: "pylav_node_connected",
            NodeDisconnectedEvent: "pylav_node_disconnected",
            NodeChangedEvent: "pylav_node_changed",
            WebSocketClosedEvent: "pylav_websocket_closed",
        }

    async def dispatch(self, event: Event):
        event_name = self.mapping[type(event)]
        self.dispatcher(event_name, event)



from __future__ import annotations

import asyncio
import contextlib
import datetime
import time
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

import asyncstdlib
import discord

from pylav._logging import getLogger
from pylav.envvars import (
    TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS,
    TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS,
)
from pylav.exceptions import EntryNotFoundError
from pylav.sql import tables
from pylav.sql.models import PlaylistModel
from pylav.types import BotT
from pylav.utils import AsyncIter

if TYPE_CHECKING:
    from pylav.client import Client

LOGGER = getLogger("PyLav.PlaylistConfigManager")


class PlaylistConfigManager:
    def __init__(self, client: Client):
        self._client = client

    @property
    def client(self) -> Client:
        return self._client

    @staticmethod
    async def get_playlist_by_name(playlist_name: str, limit: int = None) -> list[PlaylistModel]:
        if limit is None:
            playlists = await tables.PlaylistRow.select().where(
                tables.PlaylistRow.name.ilike(f"%{playlist_name.lower()}%")
            )
        else:
            playlists = (
                await tables.PlaylistRow.select()
                .where(tables.PlaylistRow.name.ilike(f"%{playlist_name.lower()}%"))
                .limit(limit)
            )

        if not playlists:
            raise EntryNotFoundError(f"Playlist with name {playlist_name} not found")
        return [PlaylistModel(**playlist) for playlist in playlists]

    @staticmethod
    async def get_playlist_by_id(playlist_id: int | str) -> PlaylistModel:
        try:
            playlist_id = int(playlist_id)
        except ValueError as e:
            raise EntryNotFoundError(f"Playlist with id {playlist_id} not found") from e
        playlist = await tables.PlaylistRow.select().where(tables.PlaylistRow.id == playlist_id).limit(1).first()
        if not playlist:
            raise EntryNotFoundError(f"Playlist with ID {playlist_id} not found")
        return PlaylistModel(**playlist)

    async def get_playlist_by_name_or_id(
        self, playlist_name_or_id: int | str, limit: int = None
    ) -> list[PlaylistModel]:
        try:
            return [await self.get_playlist_by_id(playlist_name_or_id)]
        except EntryNotFoundError:
            return await self.get_playlist_by_name(playlist_name_or_id, limit=limit)

    @staticmethod
    async def get_playlists_by_author(author: int) -> list[PlaylistModel]:
        playlists = await tables.PlaylistRow.select().where(tables.PlaylistRow.author == author)
        if not playlists:
            raise EntryNotFoundError(f"Playlist with author {author} not found")
        return [PlaylistModel(**playlist) for playlist in playlists]

    @staticmethod
    async def get_playlists_by_scope(scope: int) -> list[PlaylistModel]:
        playlists = await tables.PlaylistRow.select().where(tables.PlaylistRow.scope == scope)
        if not playlists:
            raise EntryNotFoundError(f"Playlist with scope {scope} not found")
        return [PlaylistModel(**playlist) for playlist in playlists]

    @staticmethod
    async def get_all_playlists() -> AsyncIterator[PlaylistModel]:
        for entry in await tables.PlaylistRow.select():
            yield PlaylistModel(**entry)

    @staticmethod
    async def get_external_playlists(*ids: int, ignore_ids: list[int] = None) -> AsyncIterator[PlaylistModel]:
        if ignore_ids is None:
            ignore_ids = []

        if ids:
            for entry in await tables.PlaylistRow.select().where(
                (tables.PlaylistRow.url.is_not_null())
                & (tables.PlaylistRow.id.is_in(ids))
                & (tables.PlaylistRow.id.not_in(ignore_ids))
            ):
                yield PlaylistModel(**entry)
        else:
            for entry in await tables.PlaylistRow.select().where(
                (tables.PlaylistRow.url.is_not_null()) & (tables.PlaylistRow.id.not_in(ignore_ids))
            ):
                yield PlaylistModel(**entry)

    @staticmethod
    async def create_or_update_playlist(
        id: int, scope: int, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> PlaylistModel:
        values = {
            tables.PlaylistRow.scope: scope,
            tables.PlaylistRow.author: author,
            tables.PlaylistRow.name: name,
            tables.PlaylistRow.url: url,
            tables.PlaylistRow.tracks: tracks or [],
        }
        playlist = (
            await tables.PlaylistRow.objects()
            .output(load_json=True)
            .get_or_create(tables.PlaylistRow.id == id, defaults=values)
        )
        if not playlist._was_created:
            await tables.PlaylistRow.update(values).where(tables.PlaylistRow.id == id)
        return PlaylistModel(**playlist.to_dict())

    @staticmethod
    async def delete_playlist(playlist_id: int) -> None:
        await tables.PlaylistRow.delete().where(tables.PlaylistRow.id == playlist_id)

    @staticmethod
    async def get_all_playlists_by_author(author: int) -> AsyncIterator[PlaylistModel]:
        for entry in await tables.PlaylistRow.select().where(tables.PlaylistRow.author == author):
            yield PlaylistModel(**entry)

    @staticmethod
    async def get_all_playlists_by_scope(scope: int) -> AsyncIterator[PlaylistModel]:
        for entry in await tables.PlaylistRow.select().where(tables.PlaylistRow.scope == scope):
            yield PlaylistModel(**entry)

    @staticmethod
    async def get_all_playlists_by_scope_and_author(scope: int, author: int) -> AsyncIterator[PlaylistModel]:
        for entry in await tables.PlaylistRow.select().where(
            tables.PlaylistRow.scope == scope, tables.PlaylistRow.author == author
        ):
            yield PlaylistModel(**entry)

    async def get_global_playlists(self) -> AsyncIterator[PlaylistModel]:
        for entry in await tables.PlaylistRow.select().where(tables.PlaylistRow.scope == self._client.bot.user.id):  # type: ignore
            yield PlaylistModel(**entry)

    async def create_or_update_global_playlist(
        self, id: int, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> PlaylistModel:
        return await self.create_or_update_playlist(
            id=id, scope=self._client.bot.user.id, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_user_playlist(
        self, id: int, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> PlaylistModel:
        return await self.create_or_update_playlist(
            id=id, scope=author, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_channel_playlist(
        self,
        channel: discord.abc.MessageableChannel,
        author: int,
        name: str,
        url: str | None = None,
        tracks: list[str] = None,
    ) -> PlaylistModel:
        return await self.create_or_update_playlist(
            id=channel.id, scope=channel.id, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_guild_playlist(
        self, guild: discord.Guild, author: int, name: str, url: str | None = None, tracks: list[str] = None
    ) -> PlaylistModel:
        return await self.create_or_update_playlist(
            id=guild.id, scope=guild.id, author=author, name=name, url=url, tracks=tracks
        )

    async def create_or_update_vc_playlist(
        self,
        vc: discord.channel.VocalGuildChannel,
        author: int,
        name: str,
        url: str | None = None,
        tracks: list[str] = None,
    ) -> PlaylistModel:
        return await self.create_or_update_playlist(
            id=vc.id, scope=vc.id, author=author, name=name, url=url, tracks=tracks
        )

    async def get_all_for_user(
        self,
        requester: int,
        empty: bool = False,
        *,
        vc: discord.channel.VocalGuildChannel = None,
        guild: discord.Guild = None,
        channel: discord.abc.MessageableChannel = None,
    ) -> tuple[list[PlaylistModel], list[PlaylistModel], list[PlaylistModel], list[PlaylistModel], list[PlaylistModel]]:
        """
        Gets all playlists a user has access to in a given context.

        Globals, User specific, Guild specific, Channel specific, VC specific.

        """
        global_playlists = [
            p async for p in self.get_all_playlists_by_scope(scope=self._client.bot.user.id) if (not empty or p.tracks)
        ]
        user_playlists = [p async for p in self.get_all_playlists_by_scope(scope=requester) if (not empty or p.tracks)]
        vc_playlists = []
        guild_playlists = []
        channel_playlists = []
        if vc is not None:
            vc_playlists = [p async for p in self.get_all_playlists_by_scope(scope=vc.id) if (not empty or p.tracks)]
        if guild is not None:
            guild_playlists = [
                p async for p in self.get_all_playlists_by_scope(scope=guild.id) if (not empty or p.tracks)
            ]
        if channel is not None:
            channel_playlists = [
                p async for p in self.get_all_playlists_by_scope(scope=channel.id) if (not empty or p.tracks)
            ]
        return global_playlists, user_playlists, guild_playlists, channel_playlists, vc_playlists

    async def get_manageable_playlists(
        self, requester: discord.abc.User, bot: BotT, *, name_or_id: str | None = None
    ) -> list[PlaylistModel]:
        if name_or_id:
            try:
                playlists = await self.get_playlist_by_name_or_id(name_or_id)
            except EntryNotFoundError:
                playlists = []
        else:
            try:
                playlists = [p async for p in self.get_all_playlists()]
            except EntryNotFoundError:
                playlists = []
        returning_list = []
        if playlists:
            async for playlist in AsyncIter(playlists):
                if await playlist.can_manage(requester=requester, bot=bot):
                    returning_list.append(playlist)
        return returning_list

    async def update_bundled_playlists(self, *ids: int) -> None:
        # NOTICE: Update the BUNDLED_PLAYLIST_IDS constant in the constants.py file
        with contextlib.suppress(
            asyncio.exceptions.CancelledError,
        ):
            old_time_stamp = self.client._config.next_execution_update_bundled_playlists

            self.client._config.next_execution_update_bundled_playlists = datetime.datetime.now(
                tz=datetime.timezone.utc
            ) + datetime.timedelta(days=TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS)
            await self.client._config.save()
            curated_data = {
                1: (
                    "Aikaterna's curated tracks",
                    "https://gist.githubusercontent.com/Drapersniper/cbe10d7053c844f8c69637bb4fd9c5c3/raw/playlist.pylav",
                ),
                2: (
                    "Anime OPs/EDs",
                    "https://gist.githubusercontent.com/Drapersniper/2ad7c4cdd4519d9707f1a65d685fb95f/raw/anime_pl.pylav",
                ),
            }
            id_filtered = {id: curated_data[id] for id in ids}
            if not id_filtered:
                id_filtered = curated_data
            for id, (name, url) in id_filtered.items():
                async with self._client.cached_session.get(
                    url, params={"timestamp": int(time.time())}, headers={"content-type": "application/json"}
                ) as response:
                    try:
                        data = await response.text()
                    except Exception as exc:
                        LOGGER.error(
                            "Built-in playlist couldn't be parsed - %s, report this error.", name, exc_info=exc
                        )
                        data = None
                    if not data:
                        self.client._config.next_execution_update_bundled_playlists = old_time_stamp
                        continue
                    if tracks := [t async for t in asyncstdlib.map(str.strip, data.splitlines()) if t]:
                        LOGGER.info("Updating bundled playlist - %s (%s)", name, id)
                        await self.create_or_update_global_playlist(
                            id=id, name=name, tracks=tracks, author=self._client.bot.user.id
                        )
                    else:
                        await self.delete_playlist(playlist_id=id)

    async def update_bundled_external_playlists(self, *ids: int) -> None:
        from pylav.query import Query

        with contextlib.suppress(
            asyncio.exceptions.CancelledError,
        ):
            old_time_stamp = self.client._config.next_execution_update_bundled_external_playlists
            self.client._config.next_execution_update_bundled_external_playlists = datetime.datetime.now(
                tz=datetime.timezone.utc
            ) + datetime.timedelta(days=TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS)
            await self.client._config.save()

            # NOTICE: Update the BUNDLED_PLAYLIST_IDS constant in the constants.py file
            curated_data = {
                1000001: (
                    # CYBER//
                    # Predä
                    ("https://open.spotify.com/playlist/2seaovjQuA2cMgltyLQUtd", "CYBER//")
                ),
                1000002: (
                    # PHONK//
                    # Predä
                    ("https://open.spotify.com/playlist/0rSd8LoXBD5tEBbSsbXqbc", "PHONK//")
                ),
                1000003: (  # bangers
                    # Predä
                    ("https://open.spotify.com/playlist/21trhbHm5hVgosPS1YpwSM", "bangers")
                ),
                1000004: ("https://open.spotify.com/playlist/0BbMjMQZ43vtdz7al266XH", "???"),
            }
            id_filtered = {id: curated_data[id] for id in ids}
            if not id_filtered:
                id_filtered = curated_data
            for id, (url, name) in id_filtered.items():
                track_list = []
                try:
                    LOGGER.info("Updating bundled external playlist - %s", id)
                    data = await self.client.get_tracks(await Query.from_string(url), bypass_cache=True)
                    name = data.get("playlistInfo", {}).get("name") or name
                    tracks_raw = data.get("tracks", [])
                    track_list = [t_ for t in tracks_raw if (t_ := t.get("track"))]
                except Exception as exc:
                    LOGGER.error(
                        "Built-in external playlist couldn't be parsed - %s, report this error.", name, exc_info=exc
                    )
                    data = None
                if not data:
                    self.client._config.next_execution_update_bundled_external_playlists = old_time_stamp
                    continue
                if track_list:
                    await self.create_or_update_global_playlist(
                        id=id, name=name, tracks=track_list, author=self._client.bot.user.id, url=url
                    )
                else:
                    await self.delete_playlist(playlist_id=id)

    async def update_external_playlists(self, *ids: int) -> None:
        from pylav.constants import BUNDLED_PLAYLIST_IDS
        from pylav.query import Query

        with contextlib.suppress(
            asyncio.exceptions.CancelledError,
        ):
            self.client._config.next_execution_update_external_playlists = datetime.datetime.now(
                tz=datetime.timezone.utc
            ) + datetime.timedelta(days=TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS)
            await self.client._config.save()

            async for playlist in self.get_external_playlists(*ids, ignore_ids=BUNDLED_PLAYLIST_IDS):
                try:
                    LOGGER.info("Updating external playlist - %s (%s)", playlist.name, playlist.id)
                    query = await self.client.get_tracks(
                        await Query.from_string(playlist.url),
                        bypass_cache=True,
                    )
                    tracks_raw = query.get("tracks", [])
                    track_list = [t_ for t in tracks_raw if (t_ := t.get("track"))]
                    name = query.get("playlistInfo", {}).get("name") or playlist.name
                    if track_list:
                        playlist.tracks = track_list
                        playlist.name = name
                        await playlist.save()
                except Exception as exc:
                    LOGGER.error(
                        "External playlist couldn't be updated - %s (%s), report this error.",
                        playlist.name,
                        playlist.id,
                        exc_info=exc,
                    )
